/* xstring.c - higher level string handling
 *
 * Copyright (C) 2005, 2012, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xmem.o xstr.o xstring.c -o xstring-test
 *
 * Based on ideas from glib.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <err.h>
#include <assert.h>

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XMEM_H
#include "xmem.h"
#endif

#ifndef XSTR_H
#include "xstr.h"
#endif

#ifndef XSTRING_H
#include "xstring.h"
#endif


/* Return a new xtring with the given properties.  */

xstring_t*
xstring_alloc (size_t size, size_t growsz)
{
	xstring_t *xs;

	assert (size > 0);
	assert (growsz > 0);

	xs = xnew (xstring_t, 1);
	xs->str = xmalloc (sizeof(char) * size);
	xs->size = size;
	xs->growsz = growsz;
	xs->len = 0;

	return xs;
}


/* Return a new xstring containing a copy of str. */

xstring_t*
xstring_new (const char *str)
{
	xstring_t *xs;

	if (str == NULL || *str == '\0') {
		xs = xstring_alloc (XSTRING_GROW_SIZE, XSTRING_GROW_SIZE);
		xs->str[0] = '\0';
	}
	else {
		size_t len, size;

		len = strlen (str);
		size = MAX(len + 1, XSTRING_GROW_SIZE);
		xs = xstring_alloc (size, XSTRING_GROW_SIZE);
		strcpy (xs->str, str);
		xs->len = len;
	}

	return xs;
}


/* Free the given xstring.
   If free_str is FALSE, reallocate the contained string
   and return a pointer to it. If free_str is TRUE, free
   the contained string and return NULL.  */

char*
xstring_free (xstring_t *xs, xbool_t free_str)
{
	char *str;

	assert (xs != NULL);

	if (free_str) {
		xfree (xs->str);
		xs->str = NULL;
	}
	else {
		xs->str = xrealloc (xs->str, sizeof(char) * (xs->len + 1));
	}
	str = xs->str;
	xfree (xs);

	return str;
}


static void
xstring_overflow (void)
{
	warnx ("xstring.c: overflow");
	exit (EXIT_FAILURE);
}


/* Reallocate space available to xs, adding the given amount.  */

static void
xstring_grow (xstring_t *xs, size_t amount)
{
	if (OVERFLOW(amount, xs->size)) {
		xstring_overflow ();
	}
	xs->size += amount;
	xs->str = xrealloc (xs->str, sizeof(char) * xs->size);
}


/* Append character c to the contents of xs.  */

xstring_t*
xstring_append_c (xstring_t *xs, int c)
{
	assert (xs != NULL);

	xs->str[xs->len] = c;
	xs->len += 1;
	if (xs->len == xs->size) {
		xstring_grow (xs, xs->growsz);
	}
	xs->str[xs->len] = '\0';

	return xs;
}


/* Append a copy of str to the contents of xs.  */

xstring_t*
xstring_append (xstring_t *xs, const char *str)
{
	size_t len;

	assert (xs != NULL);
	assert (str != NULL);

	len = strlen (str);
	if (len + xs->len > xs->size) {
		xstring_grow (xs, len + xs->growsz);
	}
	strcpy (&xs->str[xs->len], str);
	xs->len += len;

	return xs;
}


/* Replace the contents of xs with a copy of str.  */

xstring_t*
xstring_assign (xstring_t *xs, const char *str)
{
	size_t len;

	assert (xs != NULL);
	assert (str != NULL);

	len = strlen (str);
	if (len > xs->size) {
		xstring_grow (xs, len + xs->growsz);
	}
	strcpy (xs->str, str);
	xs->len = len;

	return xs;
}


/* Place a '\0' at the given position in xs.  */

xstring_t*
xstring_truncate (xstring_t *xs, size_t len)
{
	assert (xs != NULL);

	xs->len = MIN(xs->len, len);
	xs->str[xs->len] = '\0';

	return xs;
}


/* Compare xstrings xs1 and xs2.  */

int
xstring_cmp (xstring_t *xs1, xstring_t *xs2)
{
	assert (xs1 != NULL);
	assert (xs2 != NULL);

	return strcmp (xs1->str, xs2->str);
}


/* Compare the first n characters of xstrings xs1 and xs2.  */

int
xstring_ncmp (xstring_t *xs1, xstring_t *xs2, size_t n)
{
	assert (xs1 != NULL);
	assert (xs2 != NULL);

	return strncmp (xs1->str, xs2->str, n);
}



#if 0
xstring_t*
xstring_insert_c (xstring_t *xs, size_t pos, int c)
{
	/* XXX */

	assert (xs != NULL);
}


xstring_t*
xstring_insert (xstring_t *xs, size_t pos, const char *str)
{
	/* XXX */

	assert (xs != NULL);
}


xstring_t*
xstring_prepend_c (xstring_t *xs, int c)
{
	/* XXX */

	assert (xs != NULL);
}


xstring_t*
xstring_prepend (xstring_t *xs, const char *str)
{
	char *tmp;

	/* XXX */

	assert (xs != NULL);

	tmp = xstr_concat (str, xs->str, NULL);
	xstring_assign (xs, tmp);
	xfree (tmp);

	return xs;
}


xstring_t*
xstring_cut (xstring_t *xs, size_t pos, size_t len)
{
	char *dest, *src;

	assert (xs != NULL);

	if (pos < xs->len) {
		if (pos + len > xs->len) {
			/* XXX */
		}
		dest = &xs->str[pos];
		src = dest + len;
		memmove (dest, src, strlen (src) + 1);
		xs->len -= len;
	}

	return xs;
}
#endif



#ifdef TEST

static void
test_xstring_cmp (xstring_t *xs1, xstring_t *xs2)
{
	char c;
	int i;

	i = xstring_cmp (xs1, xs2);
	if (i < 0) {
		c = '<';
	}
	else if (i > 0) {
		c = '>';
	}
	else {
		c = '=';
	}
	printf ("%s %c %s\n", XSTRING_STR(xs1), c, XSTRING_STR(xs2));
}


int
main (int argc, char *argv[])
{
	xstring_t *xs, *xs1, *xs2;
	char *s;

	xs = xstring_new (NULL);
	xstring_append_c (xs, 'a');
	xstring_append_c (xs, 'b');
	xstring_append_c (xs, 'c');

	printf ("[%s]\n", XSTRING_STR(xs));

	xstring_free (xs, TRUE);

	xs = xstring_new ("Hello,");
	xstring_append_c (xs, ' ');
	xstring_append (xs, "world!");

	s = xstring_free (xs, FALSE);

	printf ("[%s]\n", s);

	xfree (s);

	xs = xstring_new ("replace me");
	xstring_assign (xs, "I'm the replacement text");

	printf ("[%s]\n", XSTRING_STR(xs));

	xstring_truncate (xs, 4);
	xstring_append (xs, "a test");

	printf ("[%s] (length: %d)\n", XSTRING_STR(xs), (int)XSTRING_LEN(xs));

	xstring_free (xs, TRUE);

	xs1 = xstring_new ("aaa");
	xs2 = xstring_new ("bbb");

	test_xstring_cmp (xs1, xs2);
	test_xstring_cmp (xs2, xs2);
	test_xstring_cmp (xs2, xs1);

	xstring_free (xs1, TRUE);
	xstring_free (xs2, TRUE);

#if 0
	xs = xstring_new ("foo");
	xstring_prepend (xs, "bar");
	printf ("[%s]\n", XSTRING_STR(xs));

	xstring_cut (xs, 3, 1);
	printf ("[%s]\n", XSTRING_STR(xs));

	xstring_cut (xs, 3, 2);
	printf ("[%s]\n", XSTRING_STR(xs));

	xstring_free (xs, TRUE);
#endif

	return EXIT_SUCCESS;
}
#endif

